/*!
  \file
  \brief タイピング履歴のグラフ表示

  \author Satofumi KAMIMURA

  $Id$

  \todo Legend に半透明の枠を描画する
*/

#include "HistoryGraph.h"
#include "CanvasSurface.h"
#include "TextSurface.h"
#include "Font.h"

using namespace qrk;
using namespace std;

enum {
    GridLines = 4,
    GridFontSize = 22,
};

struct HistoryGraph::pImpl
{
    CanvasSurface canvas_;
    float alpha_;

    vector<float> start_sec_data_;
    vector<float> end_sec_data_;
    vector<float> type_miss_data_;

    float end_sec_max_;
    float type_miss_max_;

    Color clear_color_;
    Color color_miss_;
    Color color_start_;
    Color color_end_;
    Color belt_color_;


    pImpl(const qrk::Rect<long>& rect)
        : canvas_(rect), alpha_(1.0), end_sec_max_(0.0), type_miss_max_(0.0),
          clear_color_(Color(0.0, 0.0, 0.0)),
          color_miss_(Color(1.0, 0.0, 0.0)), color_start_(Color(0.0, 0.6, 1.0)),
          color_end_(Color(0.0, 0.0, 1.0)), belt_color_(0.9, 0.9, 1.0, 0.6)
    {
    }


    void create(void)
    {
        // 軸のグリッドを決定する
        float left_grid_step = max(gridStep(type_miss_max_), 0.1f);
        float right_grid_step = max(gridStep(end_sec_max_), 1.0f);

        // グラフ枠の描画
        Point<long> graph_offset(GridFontSize * 2.0, GridFontSize * 1.0);
        Rect<long> graph_rect(canvas_.rect());
        graph_rect.w -= graph_offset.x * 2;
        graph_rect.h -= graph_offset.y * 2;
        graph_offset.y -= GridFontSize * 0.5;

        vector<Point<long> > points;
        points.push_back(Point<long>(graph_offset.x, graph_offset.y));
        points.push_back(Point<long>(graph_offset.x,
                                     graph_offset.y + graph_rect.h));
        points.push_back(Point<long>(graph_offset.x + graph_rect.w,
                                     graph_offset.y));
        points.push_back(Point<long>(graph_offset.x + graph_rect.w,
                                     graph_offset.y + graph_rect.h));

        // 外枠の描画
        canvas_.setColor(Color(1.0, 1.0, 3.0, 0.4));
        canvas_.drawQuadStrip(points);

        // 外枠の線を描画
        swap(points[2], points[3]);
        canvas_.setColor(Color(0.0, 0.0, 0.0));
        canvas_.drawLineLoop(points);

        int n = start_sec_data_.size();
        float x_step = (n <= 1) ? 1.0 : 1.0 * graph_rect.w / (n - 1);
        vector<Point<long> > quad_points;
        vector<Point<long> > end_points;

        // グラフ線の描画
        points.clear();
        for (int i = 0; i < n; ++i) {
            long x = graph_offset.x + (x_step * i);
            long y_base = graph_offset.y + graph_rect.h;
            long y = y_base - (graph_rect.h * start_sec_data_[i] /
                               (right_grid_step * (GridLines + 1)));
            long end_y = y_base - (graph_rect.h * end_sec_data_[i] /
                                   (right_grid_step * (GridLines + 1)));
            points.push_back(Point<long>(x, y));
            end_points.push_back(Point<long>(x, end_y));

            quad_points.push_back(Point<long>(x, y));
            quad_points.push_back(Point<long>(x, end_y));
        }

        canvas_.setColor(belt_color_);
        if (n > 1) {
            canvas_.drawQuadStrip(quad_points);
        } else if (n == 1) {
            // データが１つのみの場合は、start, end 間を直線で描画する
            quad_points[0].x += graph_rect.w / 2;
            quad_points[1].x += graph_rect.w / 2;
            points[0].x += graph_rect.w / 2;
            end_points[0].x += graph_rect.w / 2;

            canvas_.drawLineStrip(quad_points);
        }

        // グリッドの描画
        canvas_.setColor(Color(0.0, 0.0, 0.0));
        canvas_.setLineStipple(2, 0xaaaa);
        drawGridY(graph_rect, graph_offset, left_grid_step, right_grid_step);
        canvas_.setLineStipple(1, 0xffff);
        drawGridX(graph_rect, graph_offset, n);

        // グラフ線の描画
        canvas_.setLineWidth(3.0);
        canvas_.setPointSize(4.0);
        canvas_.setColor(color_start_);
        canvas_.drawLineStrip(points);
        canvas_.drawPoints(points);

        canvas_.setColor(color_end_);
        canvas_.drawLineStrip(end_points);
        canvas_.drawPoints(end_points);

        canvas_.setColor(color_miss_);
        points.clear();
        for (int i = 0; i < n; ++i) {
            long x = graph_offset.x + (x_step * i);
            long y_base = graph_offset.y + graph_rect.h;
            long y = y_base - (graph_rect.h * type_miss_data_[i] /
                               (left_grid_step * (GridLines + 1)));
            points.push_back(Point<long>(x, y));
        }
        canvas_.drawLineStrip(points);

        if (n == 1) {
            points[0].x += graph_rect.w / 2;
        }
        canvas_.drawPoints(points);

        // 線の説明の描画
        drawLegend(graph_rect, graph_offset);

        // グラフの描画が端に重なるので、線を上から描画しておく
        points.clear();
        points.push_back(graph_offset);
        points.push_back(Point<long>(points[0].x,
                                     points[0].y + graph_rect.h));
        canvas_.setLineWidth(1.0);
        canvas_.setColor(Color(0.0, 0.0, 0.0));
        canvas_.drawLineStrip(points);

        canvas_.setColor(clear_color_);
        points[0].y -= 2;
        points[1].y += 2;
        for (int i = 0; i < 2; ++i) {
            --points[0].x;
            --points[1].x;
            canvas_.drawLineStrip(points);
        }

        canvas_.setColor(Color(0.0, 0.0, 0.0));

        points.clear();
        points.push_back(Point<long>(graph_offset.x + graph_rect.w,
                                     graph_offset.y));
        points.push_back(Point<long>(points[0].x,
                                     points[0].y + graph_rect.h));
        points[0].y -= 2;
        points[1].y += 2;
        canvas_.drawLineStrip(points);

        canvas_.setColor(clear_color_);
        for (int i = 0; i < 3; ++i) {
            ++points[0].x;
            ++points[1].x;
            canvas_.drawLineStrip(points);
        }

        if (n <= 0) {
            // 表示できるデータがないときのメッセージ表示
            const char text[] = "データがありません";
            Font* font = canvas_.font();
            Rect<long> text_rect = TextSurface::surfaceSize(*font, text);
            Point<long> position(graph_offset.x +
                                 ((graph_rect.w - text_rect.w) / 2),
                                 graph_rect.h / 2);
            canvas_.drawText(text, position);
        }
    }


    void drawGridY(const Rect<long>& graph_rect,
                   const Point<long>& graph_offset,
                   float left_grid_step, float right_grid_step)
    {
        vector<Point<long> > points;
        long y_step = graph_rect.h / (GridLines + 1);
        for (size_t i = 0; i < GridLines; ++i) {
            long y_offset = y_step * (i + 1);
            points.clear();
            points.push_back(Point<long>(graph_offset.x,
                                         graph_offset.y + graph_rect.h
                                         - y_offset));
            points.push_back(Point<long>(points[0].x + graph_rect.w,
                                         points[0].y));
            canvas_.drawLineStrip(points);
        }

        // !!! フォント描画の重複をなんとかする
        enum { Offset = 5 };
        for (size_t i = 0; i < (GridLines + 1); ++i) {
            // 数値の描画
            enum {
                BufferSize = 12,
                FontSize = 17,
            };
            char buffer[BufferSize];
            if (left_grid_step < 1.0f) {
                snprintf(buffer, BufferSize, "%.1f", left_grid_step * i);
            } else {
                snprintf(buffer, BufferSize, "%.f", left_grid_step * i);
            }
            Font* font = canvas_.font();
            font->setPixelSize(FontSize);
            Rect<long> text_rect = TextSurface::surfaceSize(*font, buffer);

            // 左側の数値の描画
            Point<long> position(graph_offset.x - text_rect.w - Offset,
                                 graph_offset.y + graph_rect.h - (FontSize / 2)
                                 - (y_step * i) - 3);
            canvas_.drawText(buffer, position);

            // 右側の数値の描画
            snprintf(buffer, BufferSize, "%.f", right_grid_step * i);
            position.x = graph_offset.x + graph_rect.w + Offset;
            canvas_.drawText(buffer, position);
        }

        // 単位系の描画
        const char times_unit[] = "[回]";
        Font* font = canvas_.font();
        Rect<long> text_rect = TextSurface::surfaceSize(*font, times_unit);
        Point<long> position(graph_offset.x - text_rect.w - (Offset / 2),
                             graph_offset.y - Offset);
        canvas_.drawText(times_unit, position);

        const char sec_unit[] = "[秒]";
        text_rect = TextSurface::surfaceSize(*font, sec_unit);
        position.x = graph_offset.x + graph_rect.w + (Offset / 2);
        canvas_.drawText(sec_unit, position);
    }


    void drawGridX(const Rect<long>& graph_rect,
                   const Point<long>& graph_offset, int n)
    {
        enum { Length = 8 };
        vector<Point<long> > points;
        for (int i = 0; i < (n - 1); ++i) {
            points.clear();
            float x = 1.0 * graph_rect.w / (n - 1) * i;
            points.push_back(Point<long>(graph_offset.x + x,
                                         graph_offset.y + graph_rect.h));
            points.push_back(Point<long>(points[0].x, points[0].y - Length));
            canvas_.drawLineStrip(points);

            points[0].y = graph_offset.y;
            points[1].y = points[0].y + Length;
            canvas_.drawLineStrip(points);
        }

        // "New", "Old" の描画
        const int offset_y = 4;
        Point<long> position(graph_offset.x + graph_rect.w,
                             graph_offset.y + graph_rect.h + offset_y);
        Font* font = canvas_.font();
        font->setPixelSize(20);
        canvas_.setFont(*font);
        const char* new_text = "最近のデータ";
        Rect<long> text_rect = TextSurface::surfaceSize(*font, new_text);
        position.x -= text_rect.w;
        canvas_.drawText(new_text, position);

        position.x = graph_offset.x;
        position.y = graph_offset.y + graph_rect.h + offset_y;
        canvas_.drawText("昔のデータ", position);
    }


    void drawLegend(const Rect<long>& graph_rect,
                    const Point<long>& graph_offset)
    {
        Font* font = canvas_.font();
        font->setPixelSize(16);
        const char start_sec_title[] = "入力の開始[秒/首]";
        Rect<long> start_sec_rect =
            TextSurface::surfaceSize(*font, start_sec_title);

        const char end_sec_title[] = "入力の終了[秒/首]";
        //Rect<long> end_sec_rect =
        //TextSurface::surfaceSize(*font, end_sec_title);

        const char type_miss_title[] = "タイプミス[回/首]";
        Rect<long> type_miss_rect =
            TextSurface::surfaceSize(*font, type_miss_title);

        // 描画(左)
        enum {
            LineLength = 18,
            LineYOffset = 6,
            Offset = 4,
            FontOffset = 20,
        };
        vector<Point<long> > points;

        // 枠の描画
        Color rect_color(1.0, 1.0, 1.0, 0.7);
        canvas_.setColor(rect_color);
        points.clear();
        Rect<long> legend_rect(0, 0,
                               type_miss_rect.w + (Offset * 4) + LineLength,
                               type_miss_rect.h + (Offset * 3));
        points.push_back(Point<long>(graph_offset.x + (Offset * 1),
                                     graph_offset.y + (Offset * 1)));
        points.push_back(Point<long>(points[0].x,
                                     points[0].y + legend_rect.h));
        points.push_back(Point<long>(points[0].x + legend_rect.w,
                                     points[0].y));
        points.push_back(Point<long>(points[2].x, points[1].y));
        canvas_.drawQuadStrip(points);

        // 枠の線を描画
        swap(points[2], points[3]);
        Color rect_line_color(0.7, 0.7, 0.7, 1.0);
        canvas_.setColor(rect_line_color);
        canvas_.setLineWidth(1.0);
        canvas_.drawLineLoop(points);

        canvas_.setLineWidth(3.0);
        points.clear();
        points.push_back(Point<long>(graph_offset.x + (Offset * 3),
                                     graph_offset.y + (Offset * 3)
                                     + LineYOffset));
        points.push_back(Point<long>(points[0].x + LineLength, points[0].y));
        canvas_.setColor(color_miss_);
        canvas_.drawLineStrip(points);

        Point<long> position(points[0].x + Offset + LineLength,
                             points[0].y - LineYOffset);
        canvas_.drawText(type_miss_title, position);


        // 描画(右)
        points.clear();

        // 枠の描画
        legend_rect = Rect<long>(0, 0,
                                 start_sec_rect.w + (Offset * 4) + LineLength,
                                 (start_sec_rect.h * 2) + (Offset * 4));

        points.clear();
        points.push_back(Point<long>(graph_offset.x + graph_rect.w
                                     - legend_rect.w - Offset,
                                     graph_offset.y + (Offset * 1)));
        points.push_back(Point<long>(points[0].x,
                                     points[0].y + legend_rect.h));
        points.push_back(Point<long>(points[0].x + legend_rect.w,
                                     points[0].y));
        points.push_back(Point<long>(points[2].x, points[1].y));
        canvas_.setColor(rect_color);
        canvas_.drawQuadStrip(points);

        // 枠の線を描画
        swap(points[2], points[3]);
        canvas_.setColor(rect_line_color);
        canvas_.setLineWidth(1.0);
        canvas_.drawLineLoop(points);

        // グラフの線を描画
        points.clear();
        points.push_back(Point<long>(graph_offset.x + graph_rect.w
                                     - start_sec_rect.w - (Offset * 3)
                                     - LineLength,
                                     graph_offset.y + (Offset * 3)
                                     + LineYOffset));
        points.push_back(Point<long>(points[0].x + LineLength, points[0].y));

        canvas_.setColor(color_start_);
        canvas_.setLineWidth(3.0);
        canvas_.drawLineStrip(points);

        position.x = points[0].x + Offset + LineLength;
        position.y = points[0].y - LineYOffset;
        canvas_.drawText(start_sec_title, position);

        position.y += FontOffset;
        canvas_.drawText(end_sec_title, position);

        points[0].y += FontOffset;
        points[1].y += FontOffset;
        canvas_.setColor(color_end_);
        canvas_.drawLineStrip(points);
    }


    float gridStep(float max_value)
    {
        int base = 1;
        float steps[] = { 0.5, 1.0, 2.0, 3.0, 25.0 };
        size_t n = sizeof(steps) / sizeof(steps[0]);
        while (true) {
            for (size_t i = 0; i < n; ++i) {
                float grid_step = steps[i] * base;
                // max_value よりも少し大きめのグリッドになるようにする
                if (((GridLines + 1) * grid_step) > (max_value * 1.05)) {
                    return grid_step;
                }
            }
            base *= 10;
        }
    }
};


HistoryGraph::HistoryGraph(const qrk::Rect<long>& rect)
    : pimpl(new pImpl(rect))
{
}


HistoryGraph::~HistoryGraph(void)
{
}


qrk::Rect<long> HistoryGraph::rect(void) const
{
    return pimpl->canvas_.rect();
}


void HistoryGraph::setEventAcceptable(bool acceptable)
{
    static_cast<void>(acceptable);
    // 入力イベントは扱わないため、実装しない
}


void HistoryGraph::setAlpha(float alpha)
{
    pimpl->alpha_ = alpha;
}


float HistoryGraph::alpha(void)
{
    return pimpl->alpha_;
}


void HistoryGraph::setRotateAngle(const qrk::Angle& angle)
{
    (void)angle;
    // !!!
}


void HistoryGraph::draw(const qrk::Rect<long>& area)
{
    static_cast<void>(area);
    // !!! area を反映させる

    if (! pimpl->canvas_.isValid()) {
        return;
    }

    // 指定位置への Surface 描画
    Rect<long> draw_rect = pimpl->canvas_.rect();
    Point<long> own_position = position();
    draw_rect.x = own_position.x;
    draw_rect.y = own_position.y;

    pimpl->canvas_.draw(NULL, &draw_rect);
}


void HistoryGraph::setFont(qrk::Font& font)
{
    pimpl->canvas_.setFont(font);
}


void HistoryGraph::setBackgroundColor(const qrk::Color& color)
{
    pimpl->clear_color_ = color;
    pimpl->canvas_.setClearColor(color);
}


void HistoryGraph::addData(float start_sec, float end_sec, float type_miss)
{
    pimpl->start_sec_data_.push_back(start_sec);
    pimpl->end_sec_data_.push_back(end_sec);
    pimpl->type_miss_data_.push_back(type_miss);

    pimpl->end_sec_max_ = max(pimpl->end_sec_max_, end_sec);
    pimpl->type_miss_max_ = max(pimpl->type_miss_max_, type_miss);
}


void HistoryGraph::create(void)
{
    // グリッドへ数値を割り振る
    pimpl->create();
}
