/*!
  \file
  \brief オプション用のコンポーネント

  \author Satofumi KAMIMURA

  $Id$
*/

#include "Option.h"
#include "LayerManager.h"
#include "CallbackEvent.h"
#include "Label.h"
#include "Menu.h"
#include <vector>

using namespace qrk;
using namespace std;


struct Option::pImpl
{
    class Cursor
    {
    public:
        Label* label;
        Point<long> offset;
        bool draw_front;

        Cursor(Label* label_, Point<long> offset_, bool draw_front_)
            : label(label_), offset(offset_), draw_front(draw_front_)
        {
        }
    };

    typedef vector<Cursor> Cursors;
    typedef vector<Menu*> Components;

    Rect<long> rect_;
    bool acceptable_;
    float alpha_;
    int current_row_index_;

    LayerManager layer_manager_;
    CallbackEvent up_event_;
    CallbackEvent down_event_;
    CallbackEvent left_event_;
    CallbackEvent right_event_;

    Components components_;
    Cursors cursors_;

    Point<long> prev_cursor_;
    Point<long> cursor_;


    pImpl(const Rect<long>& rect)
        : rect_(rect), acceptable_(false), alpha_(1.0),
          current_row_index_(0),
          prev_cursor_(layer_manager_.cursor()), cursor_(prev_cursor_)
    {
        up_event_.key(SDLK_UP, SDL_KEYDOWN, KMOD_NONE);
        up_event_.key(SDLK_p, SDL_KEYDOWN, KMOD_LCTRL);
        up_event_.key(SDLK_p, SDL_KEYDOWN, KMOD_RCTRL);

        down_event_.key(SDLK_DOWN, SDL_KEYDOWN, KMOD_NONE);
        down_event_.key(SDLK_n, SDL_KEYDOWN, KMOD_LCTRL);
        down_event_.key(SDLK_n, SDL_KEYDOWN, KMOD_RCTRL);

        left_event_.key(SDLK_LEFT, SDL_KEYDOWN, KMOD_NONE);
        left_event_.key(SDLK_b, SDL_KEYDOWN, KMOD_LCTRL);
        left_event_.key(SDLK_b, SDL_KEYDOWN, KMOD_RCTRL);

        right_event_.key(SDLK_RIGHT, SDL_KEYDOWN, KMOD_NONE);
        right_event_.key(SDLK_f, SDL_KEYDOWN, KMOD_LCTRL);
        right_event_.key(SDLK_f, SDL_KEYDOWN, KMOD_RCTRL);

        layer_manager_.insertEvent(&up_event_);
        layer_manager_.insertEvent(&down_event_);
        layer_manager_.insertEvent(&left_event_);
        layer_manager_.insertEvent(&right_event_);
    }


    ~pImpl(void)
    {
        layer_manager_.removeEvent(&up_event_);
        layer_manager_.removeEvent(&down_event_);
        layer_manager_.removeEvent(&left_event_);
        layer_manager_.removeEvent(&right_event_);
    }


    void draw(const Rect<long>& area)
    {
        // マウスの移動を検出したら、キー入力用のカーソル表示を消す
        if ((prev_cursor_.x != cursor_.x) || (prev_cursor_.y != cursor_.y)) {
            removeCursor();
        }
        prev_cursor_ = cursor_;
        cursor_ = layer_manager_.cursor();

        // 上下キーの処理
        int next_row_index = current_row_index_;
        bool up = up_event_.isActive();
        bool down = down_event_.isActive();
        if (up && (current_row_index_ > 0)) {
            next_row_index = current_row_index_ - 1;
        }
        if (down &&
            (current_row_index_ < static_cast<int>(components_.size()) - 1)) {
            next_row_index = current_row_index_ + 1;
        }
        if (next_row_index != current_row_index_) {
            updateSelectedRow(next_row_index);
        }

        // 左右キーの処理
        bool left = left_event_.isActive();
        bool right = right_event_.isActive();
        if (left) {
            updateSelectedItem(current_row_index_, -1);
        }
        if (right) {
            updateSelectedItem(current_row_index_, +1);
        }

        // 各コンポーネントの描画
        int index = 0;
        for (Components::iterator it = components_.begin();
             it != components_.end(); ++it, ++index) {
            (*it)->draw(area);
        }
    }


    void updateSelectedRow(int next_row_index)
    {
        removeCursor();
        placeCursor(next_row_index);

        current_row_index_ = next_row_index;
    }


    void removeCursor(void)
    {
        Menu* menu = components_[current_row_index_];
        menu->unregisterIconLabel();
    }


    void placeCursor(int index)
    {
        Cursor& cursor = cursors_[index];
        Menu* menu = components_[index];
        menu->setIconOffset(cursor.offset);
        menu->registerIconLabel(*cursor.label, cursor.draw_front);

        // カーソルの操作を打ち消すために setIndex() を呼び出して
        // フォーカスを外す。フォーカスを外すと release() 扱いになるので、
        // press() を呼び出している
        size_t clicked_id = menu->clickedId();
        menu->setIndex(Menu::InvalidIndex);
        menu->press(clicked_id);
    }


    void updateSelectedItem(int index, int xx)
    {
        Menu* menu = components_[index];
        int max_item_index = menu->indexSize();
        int next_item_index = static_cast<int>(menu->clickedId()) + xx;
        if ((next_item_index <= 0) || (next_item_index > max_item_index)) {
            return;
        }

        menu->release();
        menu->press(next_item_index);
    }


    void addSubMenu(Menu* sub_menu, const Point<long>& position,
                    Label* cursor_label, const Point<long>& cursor_offset,
                    bool draw_front)
    {
        sub_menu->setPosition(position);
        sub_menu->setEventAcceptable(false);
        components_.push_back(sub_menu);

        cursors_.push_back(Cursor(cursor_label, cursor_offset, draw_front));

        // 最初に登録された項目にアイコンを適用する
        if (cursors_.size() == 1) {
            updateSelectedRow(0);
        }
    }
};


Option::Option(const qrk::Rect<long>& rect) : pimpl(new pImpl(rect))
{
}


Option::~Option(void)
{
}


qrk::Rect<long> Option::rect(void) const
{
    return pimpl->rect_;
}


void Option::setEventAcceptable(bool acceptable)
{
    pimpl->acceptable_ = acceptable;
}


void Option::setAlpha(float alpha)
{
    pimpl->alpha_ = alpha;
}


float Option::alpha(void)
{
    return pimpl->alpha_;
}


void Option::setRotateAngle(const qrk::Angle& angle)
{
    (void)angle;
    // !!!
}


void Option::draw(const qrk::Rect<long>& area)
{
    pimpl->draw(area);
}


void Option::addSubMenu(Menu* sub_menu, const qrk::Point<long>& position,
                        qrk::Label* cursor_label,
                        const qrk::Point<long>& cursor_offset, bool draw_front)
{
    pimpl->addSubMenu(sub_menu, position,
                      cursor_label, cursor_offset, draw_front);
}
