-- 和歌に関する処理
--
-- Satofumi KAMIMURA
-- $Id$

require("waka_data")
require("kimariji_group")
require("db_local")


-- 歴史的仮名使いを現代仮名使いに変換する
function waka_handler_convert_old_kana(text)

   local original = Utf8(text)
   local original_size = original:size()
   local converted = ""

   for i = 0, original_size do
      local ch = original:ch(i)

      if ch == Utf8("ゐ"):ch(0) then
         converted = converted .. Utf8("い"):toStdString()
      elseif ch == Utf8("ゑ"):ch(0) then
         converted = converted .. Utf8("え"):toStdString()
      else
         converted = converted .. original:substr(i, 1):toStdString()
      end
   end

   return converted
end


-- 入力された文字列が、和歌の文字列と一致しているかを返す
function waka_handler_input_is_completed(waka_id, input_ku, converted_text)

   local expected = Utf8(waka_handler_convert_old_kana(
                            waka_data_kana_[waka_id][4 + input_ku]))
   local actual = Utf8(converted_text)

   if actual:size() ~= expected:size() then
      return false
   end
   if utf8_matchSize(expected, actual) ~= expected:size() then
      return false
   end

   return true
end


-- 入力候補のローマ字文字列の生成
function waka_handler_input_sample(sample_text, current_input)

   -- 正常に入力されている部分文字列を取り除く
   local expected = Utf8(sample_text)
   local expected_size = expected:size()
   local actual = Utf8(current_input)
   local actual_size = actual:size()

   local match_size = utf8_matchSize(expected, actual)

   if match_size == expected_size then

      if actual_size > expected_size then
         return "", true
      else
         -- 入力が完全に一致しており、入力候補はない
         return "", false
      end
   end

   local left_expected = expected:substr(match_size, expected_size - match_size)
   local left_actual = actual:substr(match_size, actual_size - match_size)

   -- 残りの入力文字列から、ローマ字入力の変換候補を生成する
   local roman_text = convertToRoman(left_expected:toStdString(),
                                     left_actual:toStdString())

   -- 入力されている文字列との比較を行い、共通部分を取り除く
   local input_size = left_actual:size()
   local roman_text_size = #roman_text

   local n = min(input_size, roman_text_size)
   match_size = 0
   for i = 0, n do
      match_size = i
      if left_actual:ch(i) ~= roman_text:byte(i + 1) then
         break
      end
   end

   -- 正常に変換された部分文字列を取り除き、表示するローマ字文字列を生成する
   roman_text = roman_text:sub(match_size + 1)
   local miss_detected = (left_actual:size() ~= match_size)

   return roman_text, miss_detected
end


-- 練習用の和歌 ID を返す
function waka_handler_practice_waka(nth)

   local all_id = {}
   for i = 1, 100 do
      table.insert(all_id, i)
   end

   all_id = waka_handler_shuffle(all_id)

   local waka_id = {}
   for i = 1, nth do
      table.insert(waka_id, all_id[i])
   end

   return waka_id
end


-- 入力開始が遅い順に和歌 ID を返す
function waka_handler_slow_waka(nth)

   local id_array

   if db_local_connect() then
      id_array = db_local_slow_waka(nth)

      -- 練習を中断したときに、選択した和歌が再度選ばれないように
      -- 選択した句以外の start_delay を増やす
      db_local_update_start_delay(id_array)

      db_local_disconnect()
   else
      -- DB に接続できない場合は、乱数で選ぶ
      id_array = waka_handler_practice_waka(nth)
   end

   return id_array
end


-- 決まり字毎に和歌 ID を返す
function waka_handler_kimariji_waka(nth)

   local group_id
   if db_local_connect() then
      group_id = db_local_slow_kimariji_waka()

      -- 練習を中断したときに、選択した和歌が再度選ばれないように
      -- 選択した句以外の start_delay を増やす
      db_local_update_kimariji_start_delay(group_id)
   else
      -- DB に接続できない場合は、乱数で選ぶ
      group_id = math.ceil(math.random() * kimariji_group_id_max())
   end

   -- 決まり字の和歌を取得し、決まり字毎にシャッフルする
   local id_array = {}
   for groups = 1, #kimariji_group_[group_id] do
      local kimariji_sub_array = kimariji_group_[group_id][groups]
      local sub_id_array = {}
      for i = 1, #kimariji_sub_array do
         table.insert(sub_id_array, kimariji_sub_array[i][1])
      end

      sub_id_array = waka_handler_shuffle(sub_id_array)

      for i = 1, #sub_id_array do
         table.insert(id_array, sub_id_array[i])
      end
   end

   return id_array, group_id
end


-- 和歌 ID をシャッフルする
function waka_handler_shuffle(waka_id_array)

   local id_array = {}
   local size = #waka_id_array

   for i = 1, #waka_id_array do
      local select_id = math.ceil(math.random() * size)
      table.insert(id_array, waka_id_array[select_id])
      waka_id_array[select_id] = waka_id_array[size]
      size = size - 1
   end

   return id_array
end


-- 下の句の最初の３文字を返す
function waka_handler_kimariji_input_sample(waka_id)

   -- 下の句の最初の３文字を返す
   local input_sample_original =
      waka_handler_convert_old_kana(waka_data_kana_[waka_id][4])
   local input_sample_utf8 = Utf8(input_sample_original)

   return input_sample_utf8:substr(0, 3):toStdString()
end